﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using HeuristicLab.Common;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis.Views {
  public partial class DensityTrackbar : UserControl {

    public decimal Value {
      get { return TickToValue(trackBar.Value); }
      set { trackBar.Value = ValueToTick(value); }
    }
    public event EventHandler ValueChanged;

    public bool Checked {
      get { return radioButton.Checked; }
      set { radioButton.Checked = value; }
    }
    public event EventHandler CheckedChanged;

    public DoubleLimit Limits {
      get { return doubleLimitView.Content; }
    }
    public event EventHandler LimitsChanged;

    private IList<double> trainingValues;

    // For VisualStudio Designer
    internal DensityTrackbar() {
      InitializeComponent();
    }

    public DensityTrackbar(string name, DoubleLimit limit, IList<double> trainingValues) {
      InitializeComponent();

      this.trainingValues = trainingValues;

      doubleLimitView.Content = limit;
      doubleLimitView.Content.ValueChanged += doubleLimit_ValueChanged;

      textBox.Text = Value.ToString(CultureInfo.InvariantCulture);
      radioButton.Text = name;

      UpdateDensityChart();
    }

    public void UpdateTrainingValues(IList<double> trainingValues) {
      this.trainingValues = trainingValues;
      UpdateDensityChart();
    }

    private void UpdateDensityChart() {
      var series = chart.Series[0];
      ClearPointsQuick(series.Points);

      int numBuckets = trackBar.Maximum - trackBar.Minimum;
      var buckets = new double[numBuckets];
      var bucketSize = (Limits.Upper - Limits.Lower) / buckets.Length;
      foreach (var value in trainingValues) {
        if (bucketSize > 0.0) {
          int bucketIndex = (int)((value - Limits.Lower) / bucketSize);
          if (bucketIndex == numBuckets) {
            bucketIndex--;
          }
          buckets[bucketIndex] += 1.0;
        }
      }

      // set minimum height of all non-zero buckets on 20% of maximum
      double min = buckets.Max() / 20.0;
      for (int i = 0; i < buckets.Length; i++) {
        if (buckets[i] >= 1 && buckets[i] < min)
          buckets[i] = min;
      }

      series.Points.DataBindY(buckets);

      // Set trackbar on highest density
      double highestDensity = buckets.Max();
      var highestIndices = buckets.Select((v, i) => new { v, i }).Where(x => x.v.IsAlmost(highestDensity)).Select(x => x.i).ToList();
      trackBar.Value = highestIndices[highestIndices.Count / 2];
    }

    #region Event Handlers
    private void trackBar_ValueChanged(object sender, EventArgs e) {
      textBox.Text = Value.ToString(CultureInfo.InvariantCulture);
      OnValueChanged();
    }
    private void doubleLimit_ValueChanged(object sender, EventArgs e) {
      UpdateDensityChart();
      OnLimitsChanged();
      OnValueChanged();
    }
    private void radioButton_CheckedChanged(object sender, EventArgs e) {
      Checked = radioButton.Checked;
      trackBar.Enabled = !Checked;
      textBox.Text = Checked ? "Plotted" : Value.ToString(CultureInfo.InvariantCulture);
      OnCheckedChanged();
    }
    protected override void OnTextChanged(EventArgs e) {
      base.OnTextChanged(e);
      radioButton.Text = Text;
    }
    #endregion

    protected virtual void OnValueChanged() {
      if (ValueChanged != null)
        ValueChanged(this, EventArgs.Empty);
    }
    protected virtual void OnCheckedChanged() {
      if (CheckedChanged != null)
        CheckedChanged(this, EventArgs.Empty);
    }
    protected virtual void OnLimitsChanged() {
      if (LimitsChanged != null)
        LimitsChanged(this, EventArgs.Empty);
    }

    #region Hepers

    private decimal TickToValue(int tick) {
      return TickToValue(tick, trackBar.Maximum - trackBar.Minimum, (decimal)Limits.Lower, (decimal)Limits.Upper);
    }
    private int ValueToTick(decimal value) {
      return ValueToTick(value, trackBar.Maximum - trackBar.Minimum, (decimal)Limits.Lower, (decimal)Limits.Upper);
    }
    private static decimal TickToValue(int tick, int numberOfTicks, decimal lower, decimal upper) {
      return lower + (upper - lower) * tick / numberOfTicks;
    }
    private static int ValueToTick(decimal value, int numberOfTicks, decimal lower, decimal upper) {
      return (int)((value - lower) / ((upper - lower) / numberOfTicks));
    }

    // workaround as per http://stackoverflow.com/questions/5744930/datapointcollection-clear-performance
    private static void ClearPointsQuick(DataPointCollection points) {
      points.SuspendUpdates();
      while (points.Count > 0)
        points.RemoveAt(points.Count - 1);
      points.ResumeUpdates();
    }

    #endregion


  }
}
